from django.contrib import admin
from .models import PublicKey


@admin.register(PublicKey)
class PublicKeyAdmin(admin.ModelAdmin):
    """Admin representation of PublicKey."""

    model = PublicKey

    list_display = ('public_key_id', 'signing_algorithm', 'hashing_algorithm', 'is_active', 'public_key', 'created_at')
    ordering = ('public_key_id', 'signing_algorithm', 'hashing_algorithm', '-created_at', 'public_key', 'is_active')
    list_filter = ('signing_algorithm', 'hashing_algorithm', 'created_at', 'is_active')
    search_fields = ('public_key_id', 'signing_algorithm', 'hashing_algorithm', 'public_key')



from .crypto import HttpCavageAuthorizationVerifier, DigestHeaderVerifier


class HttpDigestRequiredMixin:
    """Force a Digest: header."""

    # should be like SHA-256=2ajR8Q+lBNm0eQW9DWWX8dZDZLB8+h0Rgmu0UCDdFrw=

    DIGEST_HEADER = 'Digest'
    CONTENT_TYPE_HEADER = 'Content-Type'
    REQUIRED_HEADERS = [
        DIGEST_HEADER,
        CONTENT_TYPE_HEADER
    ]

    def dispatch(self, request, *args, **kwargs):
        """Dispatch the object."""
        digest_verifier = DigestHeaderVerifier(request)
        digest_verifier.is_valid(raise_exceptions=True)
        request.digest_verifier = digest_verifier
        return super().dispatch(request, *args, **kwargs)


class HttpCrypoAuthorizationRequiredMixin:
    """Create a Cavage HTTP Authorization Mixin."""

    # should be Signature keyId=<key-id>,algorithm="rsa-sha256",headers="(request-target) date digest",signature=<signature-string>

    def dispatch(self, request, *args, **kwargs):
        """Dispatch the object."""
        authorization_verifier = HttpCavageAuthorizationVerifier(request, args, kwargs)
        authorization_verifier.is_valid(raise_exceptions=True)
        request.authorization_verifier = authorization_verifier
        request.public_key = None
        if hasattr(authorization_verifier, 'public_key'):
            request.public_key = authorization_verifier.public_key
        return super().dispatch(request, *args, **kwargs)



# from django.http import HttpResponse
from django.utils.translation import ugettext as _
from rest_framework.exceptions import AuthenticationFailed


class MissingHttpHeaderException(Exception):
    """Missing HTTP Header."""

    def __init__(self, header):
        """Initialize."""
        default_message = _('HTTP header required: "{}"'.format(header))
        super().__init__(default_message)


class InvalidAuthorizationTypeFormatException(Exception):
    """Invalid Authorization Type in Authorization header."""

    def __init__(self, *args, **kwargs):
        """Initialize."""
        default_message = _('Invalid authorization format')
        if not (args or kwargs):
            args = (default_message, )
        super().__init__(*args, **kwargs)


class InvalidAuthorizationTypeException(Exception):
    """Invalid Authorization in Authorization header. Should be "Signature"."""

    def __init__(self, *args, **kwargs):
        """Initialize."""
        default_message = _('Invalid authorization type')
        if not (args or kwargs):
            args = (default_message, )
        super().__init__(*args, **kwargs)


class InvalidAuthorizationKeyParameterException(Exception):
    """Invalid Authorization Key Parameter in Authorization header."""

    def __init__(self, header):
        """Initialize."""
        default_message = _('Invalid authorization key: "{}"'.format(header))
        super().__init__(default_message)


class MissingKeyIdParameterException(Exception):
    """Missing the 'keyId='' parameter in Authorization header."""

    def __init__(self, *args, **kwargs):
        """Initialize."""
        default_message = _('Could not find a "keyId" parameter')
        if not (args or kwargs):
            args = (default_message, )
        super().__init__(*args, **kwargs)


class MissingAlgorithmParameterException(Exception):
    """Missing the 'algorithm='' parameter in Authorization header."""

    def __init__(self, *args, **kwargs):
        """Initialize."""
        default_message = _('Missing algorithm')
        if not (args or kwargs):
            args = (default_message, )
        super().__init__(*args, **kwargs)


class UnsupportedAlgorithmException(Exception):
    """Unsupported algorithm parameter in Authorization header."""

    def __init__(self, algorithm):
        """Initialize."""
        default_message = _('Unsupported algorithm "{}"'.format(algorithm))
        super().__init__(default_message)


class MissingSignatureParameterException(Exception):
    """Missing the 'signature='' parameter in Authorization header."""

    def __init__(self, *args, **kwargs):
        """Initialize."""
        default_message = _('Missing signature')
        if not (args or kwargs):
            args = (default_message, )
        super().__init__(*args, **kwargs)


class DuplicateHeaderParameterException(Exception):
    """Duplicate the 'header='' parameter in Authorization header."""

    def __init__(self, header):
        """Initialize."""
        default_message = _('Duplicate authorization header "{}"'.format(header))
        super().__init__(default_message)


class MissingCavageHeaderParameterException(Exception):
    """Requested an unknown Parameter Header."""

    def __init__(self, header):
        """Initialize."""
        default_message = _('Missing cavage header parameter: "{}"'.format(header))
        super().__init__(default_message)


class EmptyCavageHeaderParameterException(Exception):
    """Requested an unknown Parameter Header."""

    def __init__(self, header):
        """Initialize."""
        default_message = _('Header parameter must not be empty')
        super().__init__(default_message)


class SignatureVerificationFailedException(AuthenticationFailed):
    """Could not verify signature."""


class NoMatchingKeysFoundException(Exception):
    """Could not verify signature."""

    def __init__(self, *args, **kwargs):
        """Initialize."""
        default_message = _('No matching keys found')
        if not (args or kwargs):
            args = (default_message, )
        super().__init__(*args, **kwargs)


class InvalidPublicKeyFormat(Exception):
    """Could not verify signature."""

    def __init__(self, *args, **kwargs):
        """Initialize."""
        default_message = _('Invalid public key format')
        if not (args or kwargs):
            args = (default_message, )
        super().__init__(*args, **kwargs)


class DigestVerificationFailedException(Exception):
    """Could not verify the digest."""

    def __init__(self, *args, **kwargs):
        """Initialize."""
        default_message = _('Digest could not be verified')
        if not (args or kwargs):
            args = (default_message, )
        super().__init__(*args, **kwargs)


class HttpResponseUnauthorized(AuthenticationFailed):
    """HTTP 401 Unauthorized Response."""



from django.db import models
from django.db.models.signals import pre_save
from .exceptions import (
    MissingHttpHeaderException,
    InvalidAuthorizationTypeFormatException,
    InvalidAuthorizationTypeException,
    InvalidAuthorizationKeyParameterException,
    MissingKeyIdParameterException,
    MissingAlgorithmParameterException,
    MissingSignatureParameterException,
    DuplicateHeaderParameterException,
    MissingCavageHeaderParameterException,
    EmptyCavageHeaderParameterException
)
from collections import OrderedDict
import hashlib
from .exceptions import (
    UnsupportedAlgorithmException,
    DigestVerificationFailedException,
)
from django.conf import settings
from django.utils.translation import ugettext as _
import base64
import uuid
import re
from Crypto.Signature import PKCS1_v1_5
from Crypto.Hash import SHA256, SHA512
from Crypto.PublicKey import RSA
from fastecdsa import curve, ecdsa
from fastecdsa.encoding.pem import PEMEncoder
from fastecdsa.encoding.der import DEREncoder


class PublicKey(models.Model):
    """Public keys."""

    SIGNING_ALGORITHM_RSA = 'RSA'
    SIGNING_ALGORITHM_ECDSA_P256 = 'EDCA-P256'
    SIGNING_ALGORITHM_ECDSA_P384 = 'EDCA-P384'
    SIGNING_ALGORITHM_ECDSA_P521 = 'EDCA-P521'
    SIGNING_ALGORITHM_ECDSA_CURVE25519 = 'EDCA-Curve25519'
    SUPPORTED_SIGNING_ALGORITHMS = [
        SIGNING_ALGORITHM_RSA,
        SIGNING_ALGORITHM_ECDSA_P256,
        SIGNING_ALGORITHM_ECDSA_P384,
        SIGNING_ALGORITHM_ECDSA_P521,
        SIGNING_ALGORITHM_ECDSA_CURVE25519,
    ]
    SIGNING_ALGORITHM_CHOICES = [
        (SIGNING_ALGORITHM_RSA, SIGNING_ALGORITHM_RSA),
        (SIGNING_ALGORITHM_ECDSA_P256, SIGNING_ALGORITHM_ECDSA_P256),
        (SIGNING_ALGORITHM_ECDSA_P384, SIGNING_ALGORITHM_ECDSA_P384),
        (SIGNING_ALGORITHM_ECDSA_P521, SIGNING_ALGORITHM_ECDSA_P521),
        (SIGNING_ALGORITHM_ECDSA_CURVE25519, SIGNING_ALGORITHM_ECDSA_CURVE25519),
    ]

    ECDSA_ALGORITHMS = [
        SIGNING_ALGORITHM_ECDSA_P256,
        SIGNING_ALGORITHM_ECDSA_P384,
        SIGNING_ALGORITHM_ECDSA_P521,
        SIGNING_ALGORITHM_ECDSA_CURVE25519,
    ]

    HASHING_ALGORITHM_SHA256 = 'SHA256'
    HASHING_ALGORITHM_SHA512 = 'SHA512'
    SUPPORTED_HASHING_ALGORITHMS = [
        HASHING_ALGORITHM_SHA256,
        HASHING_ALGORITHM_SHA512,
    ]
    HASHING_ALGORITHM_CHOICES = [
        (HASHING_ALGORITHM_SHA256, HASHING_ALGORITHM_SHA256),
        (HASHING_ALGORITHM_SHA512, HASHING_ALGORITHM_SHA512),
    ]

    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, null=True, blank=True)
    public_key_id = models.CharField(max_length=32, null=True, blank=True)
    public_key = models.TextField(max_length=500)
    hashing_algorithm = models.CharField(max_length=32, choices=HASHING_ALGORITHM_CHOICES)
    signing_algorithm = models.CharField(max_length=32, choices=SIGNING_ALGORITHM_CHOICES)
    created_at = models.DateTimeField(auto_now_add=True)
    is_active = models.BooleanField(default=False)

    def __str__(self):
        """Represent as string."""
        return str(self.public_key_id)

    def get_key_id(self):
        """Create a unique identifier for the public key."""
        return uuid.uuid4().hex

    @staticmethod
    def deflate(public_key_string):
        """Compress a public key."""
        return re.sub(r'-----(BEGIN|END) ([^ ]*) ?PUBLIC KEY-----', '', public_key_string).replace('\n', '').replace('\r', '')

    @staticmethod
    def inflate(public_key_string):
        """Add fluff to a public key if required."""
        block_length = 64
        if public_key_string.find('-----BEGIN PUBLIC KEY-----\n') != 0:
            public_key_string = '-----BEGIN PUBLIC KEY-----\n' + \
                '\n'.join(public_key_string[i:i + block_length] for i in range(0, len(public_key_string), block_length)) + \
                '\n-----END PUBLIC KEY-----'
        return public_key_string

    @property
    def deflated_public_key(self):
        """Return deflated public key."""
        return PublicKey.deflate(self.public_key)

    @property
    def inflated_public_key(self):
        """Return deflated public key."""
        return PublicKey.inflate(self.public_key)

    def validate_signature(self, verification_string, signature, charset):
        """Return True if signature is validated."""
        if self.signing_algorithm in self.ECDSA_ALGORITHMS:
            return self.__validate_fastecdsa_signature(verification_string, signature, charset)
        else:
            return self.__validate_crypto_signature(verification_string, signature, charset)

    def __validate_fastecdsa_signature(self, verification_string, signature, charset):
        """Return True if signature is valid, using FastEDSA algorithms."""
        ecdsa_public_key = PEMEncoder.decode_public_key(self.inflated_public_key)
        # signature_length = len(signature)
        r, s = DEREncoder.decode_signature(signature)
        # r_bytes, s_bytes = signature[:signature_length // 2], signature[signature_length // 2:signature_length]
        # r, s = int.from_bytes(r_bytes, 'big', signed=False), int.from_bytes(s_bytes, 'big', signed=False)
        if self.signing_algorithm == self.SIGNING_ALGORITHM_ECDSA_P256:
            curve_algorithm = curve.P256
        elif self.signing_algorithm == self.SIGNING_ALGORITHM_ECDSA_P384:
            curve_algorithm = curve.P384
        elif self.signing_algorithm == self.SIGNING_ALGORITHM_ECDSA_P521:
            curve_algorithm = curve.P2521
        elif self.signing_algorithm == self.SIGNING_ALGORITHM_ECDSA_CURVE25519:
            curve_algorithm = curve.W25519
        else:
            raise UnsupportedAlgorithmException(self.signing_algorithm)

        if self.hashing_algorithm == self.HASHING_ALGORITHM_SHA256:
            hash_function = hashlib.sha256
        elif self.hashing_algorithm == self.HASHING_ALGORITHM_SHA512:
            hash_function = hashlib.sha512
        else:
            raise UnsupportedAlgorithmException(self.hashing_algorithm)

        is_valid = False
        try:
            ecdsa.verify((r, s), signature, ecdsa_public_key, curve=curve_algorithm, hashfunc=hash_function)
            is_valid = True
        except ecdsa.EcdsaError:
            is_valid = False
        return is_valid

    def __validate_crypto_signature(self, verification_string, signature, charset):
        """Return True if signature is valid, using pycryptom algorithms."""
        rsa_public_key = RSA.import_key(self.inflated_public_key)
        signer = PKCS1_v1_5.new(rsa_public_key)

        if self.hashing_algorithm == self.HASHING_ALGORITHM_SHA256:
            digest = SHA256.new()
        elif self.hashing_algorithm == self.HASHING_ALGORITHM_SHA512:
            digest = SHA512.new()
        else:
            raise UnsupportedAlgorithmException(self.hashing_algorithm)

        digest.update(verification_string.encode(charset))
        is_valid = signer.verify(digest, signature)
        return is_valid

    @classmethod
    def pre_save(cls, sender, instance, **kwargs):
        """Sent when post_save signal is sent."""
        if instance is not None:
            instance.public_key = PublicKey.deflate(instance.public_key)
            if instance.public_key_id is None or instance.public_key_id == '':
                instance.public_key_id = instance.get_key_id()
                instance.save()


pre_save.connect(PublicKey.pre_save, sender=PublicKey)


class DigestHeader:
    """Http Digest Header."""

    DIGEST_HEADER = 'Digest'
    CONTENT_TYPE_HEADER = 'Content-Type'
    REQUIRED_HEADERS = [
        DIGEST_HEADER,
    ]

    digest = None
    charset = None
    content_type = None

    def __init__(self, request):
        """Process headers."""
        self.process(request)

    def get_missing_headers(self):
        """Returns the missing headers."""
        for header in self.REQUIRED_HEADERS:
            if header not in self.headers:
                raise MissingHttpHeaderException(_('Require header "{}"'.format(header)))

    def get_content_type_and_charset(self):
        """Get content type and charset. Default to utf-8."""
        try:
            content_type, charset_info = self.headers[self.CONTENT_TYPE_HEADER].split(";")
            content_type = content_type.lower()
        except ValueError:
            content_type = None
            charset_info = 'encoding=utf-8'
        charset_title, charset = charset_info.split("=")
        self.charset = charset.lower()
        self.content_type = content_type

    def verify(self):
        """Verify the digest header."""
        try:
            digest_algorithm, base64_digest = self.headers[self.DIGEST_HEADER].split("=", 1)
            self.digest = base64.urlsafe_b64decode(base64_digest.encode(self.charset))
        except ValueError:
            raise UnsupportedAlgorithmException(digest_algorithm)

        request_body = self.request.body
        if digest_algorithm == 'SHA-512':
            test_digest = hashlib.sha512(request_body).digest()
        elif digest_algorithm == 'SHA-384':
            test_digest = hashlib.sha384(request_body).digest()
        elif digest_algorithm == 'SHA-256':
            test_digest = hashlib.sha256(request_body).digest()
        elif digest_algorithm == 'SHA-1':
            test_digest = hashlib.sha1(request_body).digest()
        elif digest_algorithm == 'MD5':
            test_digest = hashlib.md5(request_body).digest()
        else:
            raise UnsupportedAlgorithmException(digest_algorithm)
        if test_digest != self.digest:
            raise DigestVerificationFailedException()
        return True

    def process(self, request):
        """Process headers."""
        self.request = request
        self.headers = request.headers
        self.get_missing_headers()
        self.get_content_type_and_charset()
        self.verify()


class CavageAuthorizationHeader:
    """Settings extracted from Cavage Authorization header."""

    AUTH_TYPE = 'signature'
    AUTHORIZATION_HEADER = 'Authorization'
    CONTENT_TYPE_HEADER = 'Content-Type'
    DIGEST_HEADER = 'Digest'
    SIGNATURE_CREATED_HEADER = '(created)'
    DATE_HEADER = 'Date'
    REQUIRED_HEADERS = [
        AUTHORIZATION_HEADER,
    ]

    request = None
    headers = []

    content_type = None
    charset = None
    encoding = None
    auth_info = {}
    public_key_id = None
    algorithm = None
    signature = None
    digest = None
    created = None
    expires = None

    def __init__(self, request):
        """Process headers."""
        self.process(request)

    def scrub(self, string):
        """Clean quotes from the front and end of a string."""
        if string[0] == '"' and string[-1] == '"':
            string = string[1:-1]
        return string

    def get_missing_headers(self):
        """Returns the missing headers."""
        for header in self.REQUIRED_HEADERS:
            if header not in self.headers:
                raise MissingHttpHeaderException(header)

    def get_content_type_and_charset(self):
        """Get content type and charset. Default to utf-8."""
        try:
            content_type, charset_info = self.headers[self.CONTENT_TYPE_HEADER].split(";")
            content_type = content_type.lower()
        except ValueError:
            content_type = None
            charset_info = 'encoding=utf-8'
        charset_title, charset = charset_info.split("=")
        self.charset = charset.lower()
        self.content_type = content_type

    def get_digest_header(self):
        """Get the digest header if available."""
        if self.DIGEST_HEADER in self.headers:
            self.digest = self.headers[self.DIGEST_HEADER]

    def has_digest(self):
        """Return True if a digest has been set."""
        return self.digest is not None

    def get_auth_type_and_info(self):
        """Get Authorization info."""
        try:
            auth_type, parameters_string = self.headers[self.AUTHORIZATION_HEADER].split(' ', 1)
            auth_type = auth_type.lower()
        except ValueError:
            raise InvalidAuthorizationTypeFormatException()
        if auth_type != 'signature':
            raise InvalidAuthorizationTypeException()
        self.auth_type = auth_type.lower()
        return auth_type, parameters_string

    def get_request_target(self):
        """Get the request target."""
        request_target = '{} {}'.format(self.request.method, self.request.path)
        querystring = self.request.GET.urlencode()
        if len(querystring) > 0:
            request_target += '?{}'.format(querystring)
        return request_target.lower()

    def parse_signature_parameters(self, parameters_string):
        """Get signature parameters."""
        try:
            auth_info_keypairs = parameters_string.split(',')
        except ValueError:
            raise InvalidAuthorizationKeyParameterException()
        signature_parameters = {}
        for auth_info_keypair in auth_info_keypairs:
            key, value = auth_info_keypair.strip().split("=", 1)
            if value[0] == '"' and value[-1] == '"':
                value = value[1:-1]
            signature_parameters[key] = value
        self.signature_parameters = signature_parameters

        # TODO: implement 'header' checks
        # must throw errors in certain conditions
        # https://tools.ietf.org/html/draft-cavage-http-signatures-12

        if 'keyId' not in signature_parameters:
            raise MissingKeyIdParameterException()
        self.public_key_id = self.scrub(signature_parameters['keyId'])

        if 'algorithm' not in signature_parameters:
            raise MissingAlgorithmParameterException()
        self.algorithm = self.scrub(signature_parameters['algorithm'])

        if 'signature' not in signature_parameters:
            raise MissingSignatureParameterException()
        self.signature = self.scrub(signature_parameters['signature'])

        if 'created' in signature_parameters:
            self.created = self.scrub(signature_parameters['created'])
        if 'expires' in signature_parameters:
            self.created = self.scrub(signature_parameters['expires'])

        if 'headers' in signature_parameters:
            # loop through header requirements
            self.get_verification_headers(self.scrub(signature_parameters['headers']))
        else:
            cleaned_created_header = self.SIGNATURE_CREATED_HEADER[1:-1]
            if cleaned_created_header in signature_parameters:
                self.verification_headers = {
                    self.SIGNATURE_CREATED_HEADER: self.signature_parameters[cleaned_created_header]
                }
            elif self.DATE_HEADER in self.headers:
                self.verification_headers = {
                    self.DATE_HEADER: self.headers[self.DATE_HEADER]
                }
            else:
                raise MissingCavageHeaderParameterException(self.SIGNATURE_CREATED_HEADER)

    def get_verification_headers(self, header_string):
        """Process the verification header requirements."""
        required_headers = header_string.strip().split(' ')
        if len(required_headers) == 0:
            raise EmptyCavageHeaderParameterException()
        verification_headers = OrderedDict()
        for required_header_lower in required_headers:
            if required_header_lower in verification_headers:
                raise DuplicateHeaderParameterException(required_header_lower)
            if required_header_lower[0] != '(':
                required_header = required_header_lower.replace('-', ' ').title().replace(' ', '-')
                if required_header not in self.headers:
                    raise MissingHttpHeaderException(required_header)
                else:
                    verification_headers[required_header_lower] = self.headers[required_header]
            else:
                required_header = required_header_lower[1:-1]
                if required_header_lower == '(request-target)':
                    verification_headers[required_header_lower] = self.get_request_target()
                else:
                    if required_header not in self.signature_parameters:
                        raise MissingCavageHeaderParameterException(required_header)
                    else:
                        verification_headers[required_header_lower] = self.signature_parameters[required_header]
        self.verification_headers = verification_headers

    def get_verification_string(self):
        """Merge required headers into verification string."""
        output_rows = []
        for key, value in self.verification_headers.items():
            output_rows.append('{}: {}'.format(key.lower(), value))
        return '\n'.join(output_rows)

    def process(self, request):
        """Process headers."""
        self.request = request
        self.headers = request.headers
        self.get_missing_headers()

        self.get_content_type_and_charset()
        auth_type, parameters_string = self.get_auth_type_and_info()
        self.get_digest_header()
        self.parse_signature_parameters(parameters_string)



import re
from pathlib import Path
import setuptools


def get_long_description():
    """Get README."""
    long_description = ''
    with open("README.md", "r") as fh:
        long_description = fh.read()
    return long_description


def get_version(package):
    """Get current version."""
    version = (Path("src") / package / "__version__.py").read_text()
    match = re.search("__version__ = ['\"]([^'\"]+)['\"]", version)
    assert match is not None
    return match.group(1)


setuptools.setup(
    name="drf-keypair-permissions",
    version=get_version('keypair_permissions'),
    author="Adonis Gaitatzis",
    author_email="backupbrain@gmail.com",
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    description="Crypto Keypair Authorization for Django Rest Framework",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/backupbrain/drf-keypair-permissions",
    project_urls={
        "Documentation": "https://drf-keypair-permissions.readthedocs.io/"
    },
    install_requires=[
        'pycryptodome',
        'fastecdsa',
    ],
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Environment :: Web Environment",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",

    ],
    python_requires=">=3.6",
)



import base64
from .models import CavageAuthorizationHeader, DigestHeader
from .exceptions import (
    UnsupportedAlgorithmException,
    SignatureVerificationFailedException,
    NoMatchingKeysFoundException,
    HttpResponseUnauthorized,
)
from .models import PublicKey


class DigestHeaderVerifier:
    """Verify a Digest."""

    in_verbose_mode = False
    raise_exceptions = True

    def __init__(self, request, *args, **kwargs):
        """Initialize class."""
        self.say("In verbose mode")
        self.request = request
        self.args = args
        self.kwargs = kwargs

    def handle_error(self, exception):
        """Return False or raise exception."""
        if self.raise_exceptions is True:
            raise exception
        else:
            return False

    def is_valid(self, raise_exceptions=True):
        """Dispatch the object."""
        try:
            self.digest_header = DigestHeader(self.request)
        except Exception as e:
            return self.handle_error(e)

    def say(self, message):
        """Print debugging messages."""
        if self.in_verbose_mode is True:
            print(message)


class HttpCavageAuthorizationVerifier:
    """Verify HTTP signatures."""

    in_verbose_mode = False
    raise_exceptions = True
    cavage_authorization_header = None
    charset = None
    public_key = None
    signature = None
    verification_string = None

    CONTENT_TYPE_HEADER = 'Content-Type'

    HEADER_ALGORITHM_HS2019 = 'hs2019'
    HEADER_ALGORITHM_RSA_SHA256 = 'rsa-sha256'
    HEADER_ALGORITHM_RSA_SHA512 = 'rsa-sha512'
    SUPPORTED_HEADER_ALGORITHMS = [
        HEADER_ALGORITHM_HS2019,
        HEADER_ALGORITHM_RSA_SHA256,
        HEADER_ALGORITHM_RSA_SHA512,
    ]

    # hs2019, --rsa-sha1--, --rsa-sha256--, --hmac-sha256--, --ecdsa-sha256--
    # hs2019 with SHA-512 hash
    def __init__(self, request, *args, **kwargs):
        """Initialize class."""
        self.say("In verbose mode")
        self.request = request
        self.args = args
        self.kwargs = kwargs

    def handle_error(self, exception):
        """Return False or raise exception."""
        if self.raise_exceptions is True:
            raise exception
        else:
            return False

    '''
    def get_rsa256_b64_message_digest(self, request_body, charset):
        """Create an RSA256 message digest."""
        expected_digest = hashlib.sha256(request_body).digest()
        expected_base64_digest = base64.b64encode(expected_digest).decode(charset)
        return expected_base64_digest
    '''

    def get_content_type_and_charset(self):
        """Get content type and charset. Default to utf-8."""
        content_type = None
        try:
            content_type, charset_info = self.request.headers[self.CONTENT_TYPE_HEADER].lower().split(";")
        except ValueError:
            message = "Could not read charset info"
            self.say("    " + message)
            charset_info = 'encoding=utf-8'
        charset_title, charset = charset_info.split("=")
        return content_type, charset

    def decode_signature(self, base64_signature, charset):
        """Base64-decode the signature."""
        self.say("    signature: {}".format(base64_signature))
        try:
            signature = base64.b64decode(base64_signature.encode(charset))
        except LookupError:
            self.say("Could not decode base64 signature")
            return self.handle_error(SignatureVerificationFailedException())
        return signature

    def get_public_key_from_id(self, public_key_id):
        """Fetch the public key from the database."""
        try:
            public_key = PublicKey.objects.get(public_key_id=public_key_id, is_active=True)
            return public_key
        except PublicKey.DoesNotExist:
            return self.handle_error(NoMatchingKeysFoundException())

    def is_valid(self, raise_exceptions=True):
        """Dispatch the object."""
        try:
            self.cavage_authorization_header = CavageAuthorizationHeader(self.request)
        except Exception as e:
            return self.handle_error(e)
        algorithm = self.cavage_authorization_header.algorithm.lower()
        if algorithm not in self.SUPPORTED_HEADER_ALGORITHMS:
            raise UnsupportedAlgorithmException(algorithm)

        base64_signature = self.cavage_authorization_header.signature
        public_key_id = self.cavage_authorization_header.public_key_id
        content_type, charset = self.get_content_type_and_charset()

        signature = self.decode_signature(base64_signature, charset)
        public_key = self.get_public_key_from_id(public_key_id)

        verification_string = self.cavage_authorization_header.get_verification_string()

        self.charset = charset
        self.public_key = public_key
        self.signature = signature
        self.verification_string = verification_string

        try:
            does_signature_validate = self.public_key.validate_signature(verification_string, signature, charset)
        except Exception as e:
            return self.handle_error(e)
        if does_signature_validate:
            return True
        else:
            return self.handle_error(SignatureVerificationFailedException())

        # if there is a digest, we need to verify that
        does_digest_match = False
        if self.cavage_authorization_header.has_digest():
            try:
                self.digest_header = DigestHeader(self.request)
                does_digest_match = True
            except Exception as e:
                self.handle_error(e)

        if does_signature_validate and does_digest_match:
            return True
        else:
            return self.handle_error(HttpResponseUnauthorized())

    def say(self, message):
        """Print debugging messages."""
        if self.in_verbose_mode is True:
            print(message)
